#include "window.h"

#include "pins.h"

#include <assert.h>
#include <librsvg/rsvg.h>
#include <pango/pangocairo.h>
#include <wiringPi.h>

/* TYPES!!! */

struct _GameWindow {

	GtkWindow parent;

	GtkDrawingArea* area;

	RsvgHandle* logo;

	enum {
		SPLASH,
		WAIT,
		LIT,
		DONE,
	} state;

	struct {
		int team_1;
		int team_2;
	} score;

	enum { RED, GREEN } colour;

	bool g1_bouncing;
	bool r1_bouncing;
	bool g2_bouncing;
	bool r2_bouncing;

};

G_DEFINE_TYPE(
	GameWindow, game_window, GTK_TYPE_WINDOW );

enum colour {
	WHITE, UNIVERSITY_BLUE,
};

/* PROTOTYPES!!! */

static void game_window_class_init( GameWindowClass* class );
static void game_window_init( GameWindow* window );

static void realise( GtkWidget* widget, void* data );

static void draw( GtkDrawingArea* area, cairo_t* cr,
	int width, int height, void* data );

static void set_cario_colour( cairo_t* cr, enum colour colour );

static int bounce_reset( void* data );
static int light( void* data );

/* PUBLIC FUNK!!! */

GameWindow* game_window_new( GameApp* app ) {

	return g_object_new( GAME_WINDOW_TYPE,
		"application", app, NULL );

}

void game_window_button_press( GameWindow* window, int pin ) {

	bool* bouncing = NULL;

	switch ( pin ) {

		case BUTTON_G1:
			bouncing = &window->g1_bouncing;
			break;

		case BUTTON_R1:
			bouncing = &window->r1_bouncing;
			break;

		case BUTTON_G2:
			bouncing = &window->g2_bouncing;
			break;

		case BUTTON_R2:
			bouncing = &window->r2_bouncing;
			break;

	}

	assert( NULL != bouncing );

	if ( *bouncing ) {
		return;
	}

	fprintf( stderr, "Ping %d\n", pin );

	*bouncing = true;

	g_timeout_add( 200, bounce_reset, bouncing );

	if ( !digitalRead( BUTTON_G1 ) && !digitalRead( BUTTON_R1 ) &&
			!digitalRead( BUTTON_G2 ) && !digitalRead( BUTTON_R2 ) ) {

		fprintf( stderr, "All pressed, reset\n" );

		window->state = SPLASH;
		window->score.team_1 = 0;
		window->score.team_2 = 0;
		gtk_widget_queue_draw( GTK_WIDGET( window->area ) );
		return;

	}

	switch ( window->state ) {

		case SPLASH: {

			fprintf( stderr, "Splash -> wait\n" );

			window->state = WAIT;
			window->score.team_1 = 0;
			window->score.team_2 = 0;
			g_timeout_add( g_random_int_range( 1000, 5000 ),
				light, window );
			break;

		}

		case WAIT: {
			break;
		}

		case LIT: {

			switch ( pin ) {

				case BUTTON_G1:
				case BUTTON_R1: {

					fprintf( stderr, "Team 1 press, " );
					if ( ( GREEN == window->colour ) ==
							( BUTTON_G1 == pin ) ) {
						fprintf( stderr, "correct\n" );
						window->score.team_1++;
						digitalWrite( LED_G1, LOW );
						digitalWrite( LED_R1, LOW );
						digitalWrite( LED_G2, LOW );
						digitalWrite( LED_R2, LOW );
						digitalWrite( LED_G3, LOW );
						digitalWrite( LED_R3, LOW );
						digitalWrite( LED_G4, LOW );
						digitalWrite( LED_R4, LOW );

						window->state = WAIT;

					} else {
						fprintf( stderr, "wrong\n" );
					}

					break;

				}

				case BUTTON_G2:
				case BUTTON_R2: {

					fprintf( stderr, "Team 2 press, " );
					if ( ( GREEN == window->colour ) ==
							( BUTTON_G2 == pin ) ) {

						fprintf( stderr, "correct\n" );
						window->score.team_2++;
						digitalWrite( LED_G1, LOW );
						digitalWrite( LED_R1, LOW );
						digitalWrite( LED_G2, LOW );
						digitalWrite( LED_R2, LOW );
						digitalWrite( LED_G3, LOW );
						digitalWrite( LED_R3, LOW );
						digitalWrite( LED_G4, LOW );
						digitalWrite( LED_R4, LOW );

						window->state = WAIT;

					} else {
						fprintf( stderr, "wrong\n" );
					}

					break;

				}

			}


			if ( window->score.team_2 >= 10 ||
					window->score.team_1 >= 10 ) {

				window->state = DONE;

			}

			if ( window->state == WAIT ) {

				g_timeout_add( g_random_int_range( 1000, 5000 ),
					light, window );

			}

			break;

		}

		case DONE: {

			break;

		}

	}

	gtk_widget_queue_draw( GTK_WIDGET( window->area ) );

}

/* PRIVATE FUNK!!! */

static void game_window_class_init( GameWindowClass* class ) {

	gtk_widget_class_set_template_from_resource(
		GTK_WIDGET_CLASS( class ), "/xyz/tomg/game/res/window.ui" );
	gtk_widget_class_bind_template_child(
		GTK_WIDGET_CLASS( class ), GameWindow, area );

	gtk_widget_class_bind_template_callback_full(
		GTK_WIDGET_CLASS( class ), "realise",
		G_CALLBACK( &realise ) );

}

static void game_window_init( GameWindow* window ) {

	gtk_widget_init_template( GTK_WIDGET( window ) );

	GBytes* logo_svg = g_resources_lookup_data(
		"/xyz/tomg/game/res/logo.svg", 0, NULL );
	GError* error = NULL;
	const char* data = g_bytes_get_data( logo_svg, NULL );
	window->logo = rsvg_handle_new_from_data(
		(const unsigned char*) data, strlen( data ), &error );
	if ( NULL != error ) {
		fprintf( stderr, "%s\n", error->message );
	}
	g_bytes_unref( logo_svg );

	window->state = SPLASH;

	gtk_drawing_area_set_draw_func(
		window->area, &draw, (void*) window, NULL );

	gtk_widget_set_cursor( GTK_WIDGET( window ),
		gdk_cursor_new_from_name( "none", NULL ) );

}

static void realise( GtkWidget* widget, void* data ) {

	gtk_window_fullscreen( GTK_WINDOW( widget ) );
	(void) data;

}

static void draw( GtkDrawingArea* area, cairo_t* cr,
		int width, int height, void* data ) {

	(void) area;
	GameWindow* window = (GameWindow*) data;

	set_cario_colour( cr, UNIVERSITY_BLUE );
	cairo_paint( cr );

	PangoFontDescription* desc = pango_font_description_new();
	pango_font_description_set_family( desc, "Roboto" );
	pango_font_description_set_absolute_size( desc, PANGO_SCALE * (
		0.8 * height < 0.66 * width ? 0.8 * height : 0.66 * width ) );

	switch ( window->state ) {

		case SPLASH: {

			rsvg_handle_render_document( window->logo,
				cr, &(RsvgRectangle) { 0, 0, width, height, }, NULL );
			break;

		}

		case WAIT:
		case LIT: {

			char buffer[5];
			snprintf( buffer, sizeof buffer,
				"%d", window->score.team_1 );

			int lw, lh;

			PangoLayout* layout = pango_cairo_create_layout( cr );
			pango_layout_set_font_description( layout, desc );

			pango_layout_set_text( layout, buffer, -1 );
			pango_layout_get_pixel_size( layout, &lw, &lh );
			fprintf( stderr, "Layout %s size %d, %d\n", buffer, lw, lh );

			cairo_move_to( cr, 0.25 * width - lw / 2.,
				0.5 * height - lh / 2. );
			set_cario_colour( cr, WHITE );
			pango_cairo_show_layout( cr, layout );

			snprintf( buffer, sizeof buffer,
				"%d", window->score.team_2 );

			pango_layout_set_text( layout, buffer, -1 );
			pango_layout_get_pixel_size( layout, &lw, &lh );

			cairo_move_to( cr, 0.75 * width - lw / 2.,
				0.5 * height - lh / 2. );
			set_cario_colour( cr, WHITE );
			pango_cairo_show_layout( cr, layout );

			g_object_unref( G_OBJECT( layout ) );

			rsvg_handle_render_document( window->logo, cr,
				&(RsvgRectangle) {
					0.4 * width, 0.8 * height,
					0.2 * width, 0.2 * height,
				}, NULL );

			break;

		}

		case DONE: {

			pango_font_description_set_absolute_size( desc,
				PANGO_SCALE * ( 0.4 * height < 0.33 * width ?
					0.2 * height : 0.16 * width ) );

			PangoLayout* layout = pango_cairo_create_layout( cr );
			pango_layout_set_font_description( layout, desc );

			pango_layout_set_text( layout,
				window->score.team_1 >= window->score.team_2 ?
					"Team 1 Wins!" : "Team 2 Wins!", -1 );
			pango_layout_set_width( layout, PANGO_SCALE* 0.8 * width );
			//pango_layout_set_alignment( layout, PANGO_ALIGN_CENTER );

			int lw, lh;
			pango_layout_get_pixel_size( layout, &lw, &lh );

			cairo_move_to( cr, 0.5 * width - lw / 2,
				0.5 * height - lh / 2 );
			set_cario_colour( cr, WHITE );
			pango_cairo_show_layout( cr, layout );

			break;

		}

	}

	pango_font_description_free( desc );

}

static void set_cario_colour( cairo_t* cr, enum colour colour ) {

	switch ( colour ) {

		case WHITE:
			cairo_set_source_rgba( cr, 1., 1., 1., 1. );
			break;

		case UNIVERSITY_BLUE:
			cairo_set_source_rgba(
				cr, 0, 0x5C / 255., 0x84 / 255., 1. );
			break;

	}

}

static int bounce_reset( void* data ) {

	*( (bool*) data ) = false;

	return G_SOURCE_REMOVE;

}

static int light( void* data ) {

	GameWindow* window = (GameWindow*) data;

	if ( WAIT != window->state ) {
		return G_SOURCE_REMOVE;
	}

	static const int leds[] = {
		LED_G1, LED_G2, LED_G3, LED_G4,
		LED_R1, LED_R2, LED_R3, LED_R4,
	};

	int led = g_random_int_range( 0, sizeof leds / sizeof *leds );
	fprintf( stderr, "Lighting led %d (pin %d)\n",
		led, leds[led] );
	for ( int i = 0; i < 8; i++ ) {
		digitalWrite( leds[i], i == led ? HIGH : LOW );
	}

	window->state = LIT;
	window->colour = led < 4 ? GREEN : RED;

	return G_SOURCE_REMOVE;

}
